# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Module for mathematical functions and utilities.

This module contains in particular:

  * Vector classes.
  * Geometry classes.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
from __future__ import absolute_import

from .kinematics import *
from .vectors import *
from .geometry import *
